"""
为各个模块提供抽象类别
"""
from damei.nn.api.utils import Config


class AbstractModule(object):
    name = 'default_name'
    status = 'stopped'
    description = 'default_description'
    default_cfg = None

    def __init__(self, default_cfg=None, cfg=None, **kwargs):
        """
        初始化模块，合并配置文件
        """
        AbstractModule.status = 'ready'
        self.name = AbstractModule.name
        self.status = AbstractModule.status
        self.description = AbstractModule.description
        self._cfg = self._init_cfg(default_cfg, cfg, **kwargs)
        self.mi = None
        self.mo = None

    def _init_cfg(self, default_cfg, cfg, **kwargs):
        """
        合并配置文件
        """
        if default_cfg is not None:
            default_cfg = Config.fromfile(default_cfg) if isinstance(default_cfg, str) else default_cfg

        if cfg is not None:
            cfg = Config.fromfile(cfg) if isinstance(cfg, str) else cfg
        cfg = default_cfg.merge(cfg2=cfg)
        return cfg

    @property
    def cfg(self):
        return self._cfg

    def train(self, *args, **kwargs):
        pass

    def infer(self, *args, **kwargs):
        pass

    def evaluate(self, *args, **kwargs):
        pass






import os, sys
from pathlib import Path
from easydict import EasyDict
import functools
import collections
import copy
import re
from . import misc


class PyConfig(collections.UserDict):
	def __init__(self, cfg_file=None, name=None, root_path=None):
		super(PyConfig, self).__init__()
		self.root_path = root_path
		# self.root_path = root_path if root_path else f'{pydir.parent.parent}'  # 上上级路径
		self._name = name if name else f'{cfg_file}'  # 其实是path
		self._items = dict()
		self.init_config(cfg_file)  # 根据配置文件把内容和属性注册到items里

		self.check_items()  # TODO

	def __len__(self):
		return len(self._items)

	def __dict__(self):
		return self._items

	def __getitem__(self, item):
		return self._items[item]

	def __repr__(self):
		format_str = f'<class {self.__class__.__name__}> ' \
					 f'(path="{self._name}", ' \
					 f'items={self.items}'
		# f'keys={self._items.keys()})'

		return format_str

	def __missing__(self, key):
		if isinstance(key, str):
			raise KeyError(key)
		return self[str(key)]

	def __contains__(self, key):
		return str(key) in self.data

	def __setitem__(self, key, item):
		self.data[str(key)] = item

	# def __getattr__(self, key):
	#    return self.data[str(key)]

	def info(self):
		"""查看当前所有配置信息"""
		# info_str = f'{self.__repr__()}\n'
		info_str = f'Configs:\n'
		info_dict = self.items
		info_str += misc.dict2info(info_dict)
		return info_str

	def init_config(self, cfg_file):
		if cfg_file is None:
			return
		cp = Path(os.path.abspath(cfg_file))  # config path
		# 根据sys.path的环境路径和cp的绝对路径，获取module_dir
		valid_sys_path = [x for x in sys.path if x in f'{cp}']
		if len(valid_sys_path) >= 1:
			rp = valid_sys_path[0]  # root_path
		else:
			sys.path.append(f'{cp.parent}')
			rp = f'{cp.parent}'
		# 提取模块路径：如：modules.detection.seyolov5.config
		module_dir = f'{str(cp.parent).replace(f"{rp}/", "").replace("/", ".")}'
		# print(module_dir)
		# print(f'rp: {rp} \ncp: {cp}\nm dir: {module_dir}')
		code = f'from {module_dir} import {cp.stem}'
		# print(cfg_file, code)
		exec(code)
		cfg = eval(cp.stem)  # 模块名

		# 读取文件获取顶格写并且不以#开头的属性
		with open(cfg_file, 'r') as f:
			data = f.readlines()
		data = [x for x in data if not (x.startswith(' ') or x.startswith('#'))]  # 非顶格的不要，# 开头的不要
		data = '\n'.join(data)
		attrs = re.findall(pattern=r"\s[a-z]\w+ ?= ?", string=data)
		attrs = [x.split('=')[0].replace('\n', '').strip() for x in attrs]

		# 注册属性到_items里，注册属性到self的属性里。
		for attr in attrs:
			if hasattr(cfg, attr):
				exec(f"self._items[attr] = cfg.{attr}")
				attr_value = self._items[attr]
				if isinstance(attr_value, dict):
					setattr(self, attr, EasyDict(self._items[attr]))
				else:
					setattr(self, attr, attr_value)
			else:
				pass

	def check_items(self):
		"""初始化末，检查所有配置项，处理：缩略语~转为真实路径。
		递归方法，如果v类型是str,判断替换，如果循环完了，结束，如果v的类型是dict，继续循环
		"""
		# print(self._items)
		for k, v in self._items.items():
			new_v = self.recursive_func(v)
			self._items[k] = new_v
			# 更新属性
			if hasattr(self, k):
				delattr(self, k)
			setattr(self, k, EasyDict(new_v)) if isinstance(new_v, dict) else setattr(self, k, new_v)

	def recursive_func(self, value):
		"""递归函数,如果v类型是str,判断替换，如果循环完了，结束，如果v的类型是dict，继续循环"""
		if value is None:
			return None
		elif isinstance(value, str):
			return value.replace('~', os.environ['HOME'])
		elif isinstance(value, int):
			return value
		elif isinstance(value, float):
			return value
		elif isinstance(value, bool):
			return value
		elif isinstance(value, list):
			new_list = [self.recursive_func(x) for x in value]
			return new_list
		elif isinstance(value, dict):
			new_dict = {}
			for k, v in value.items():
				new_dict[k] = self.recursive_func(v)
			return new_dict
		else:
			raise TypeError(f'Config file only support basic python types but the value type is: {type(value)}.')

	def update_item(self, attrs, value):
		"""更新配置条目
		attrs: 长度为0时代表，更新全部value，长度为1时，代表第1个元素是第一层，
		"""
		if len(attrs) == 0:
			self.merge(value)
		elif len(attrs) == 1:
			self._items[attrs[0]] = value
		elif len(attrs) == 2:
			self._items[attrs[0]][attrs[1]] = value
		elif len(attrs) == 3:
			self._items[attrs[0]][attrs[1]][attrs[2]] = value
		else:
			self._items[attrs[0]][attrs[1]][attrs[2]][attrs[3]] = value
		self.check_items()

	def merge(self, cfg2):
		"""合并另一个配置文件到内部，是inplace的"""
		if cfg2 is None:
			return self
		for k2, v2 in cfg2._items.items():
			merged_v = self.recursive_func2(v2, k2=k2, sub_items=self._items)
			self._items[k2] = merged_v
		# print(self.items)
		self.check_items()
		return self

	def recursive_func2(self, v2, k2=None, sub_items=None):
		"""v1是之前的值，v2是现在的值"""
		if k2 and (k2 not in list(sub_items.keys())):
			return v2
		else:
			v1 = sub_items[k2]

			if type(v1) != type(v2):
				return v2
			f = functools.partial(isinstance, v1)
			if f(str) or f(int) or f(float) or f(bool):
				return v2
			elif f(list):
				tmp = [x for x in v2 if x not in v1]
				return v1 + tmp
			elif f(dict):  # v1v2都是dict
				new_dict = v1  # 保留v1
				for kk, vv in v2.items():
					new_dict[kk] = self.recursive_func2(vv, k2=kk, sub_items=v1)
				return new_dict

	@property
	def items(self):
		return self._items

	def keys(self):
		return self._items.keys()

	def get(self, key, default=None):
		if key in list(self._items.keys()):
			return self._items[key]
		else:
			return default



"""
damei scp 配置好目标机器，同步文件
usage:
    dmscp siyuan ./xxx.zip  # 默认是上传 -u --upload, siyuan是配置好的服务器
    dmscp siyuan -d ./xxx.zip -d # 下载 -d --download

"""
import os, sys
import argparse

class RemoteServer(object):
    """远程服务器"""
    def __init__(self, host, port, user, password):
        self.host = host
        self.port = port
        self.user = user
        self.password = password


class DmSCP(object):
    def __init__(self):
        pass

    def upload(self, src, dst):
        pass

    def download(self, src, dst):
        pass


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='damei scp')
    parser.add_argument('server', help='server name')
    pass









"""
提供基本的UAII类，把各种函数封装在这里。
"""
import os, sys
import numpy as np
import damei as dm
import cv2
logger = dm.getLogger('base_uaii')
from . import general


class BaseUAII(object):
    def __init__(self):
        self._modules = None
        self._scripts = None
        self._ios = None
        self._streams = None
        self._name2cls = None
        self.cfg = None

    def _configure(self):
        """初始化配置，还需要考虑重新初始化配置"""
        streams = self.streams
        for i, (k, v) in enumerate(self.streams.module_dict.items()):
             print(i, k, isinstance(v, type))
        for i, (k, v) in enumerate(self.modules.module_dict.items()):
            print(i, k, type(v), v, isinstance(v, type))
        exit()
        pass

    def register_name2cls(self):
        name2cls = dict()
        for k, v in self.modules.module_dict.items():
            name = k
            name2cls[name] = 'MODULE'
        for k, v in self.scripts.module_dict.items():
            name2cls[k] = 'SCRIPT'
        for k, v in self.streams.module_dict.items():
            name2cls[k] = 'STREAM'
        for k, v in self.ios.module_dict.items():
            name2cls[k] = 'IO'
        return name2cls

    @property
    def modules(self):
        return self._modules

    @property
    def scripts(self):
        return self._scripts

    @property
    def streams(self):
        return self._streams

    @property
    def ios(self):
        return self._ios

    def get_module(self, name, cls=None, *args, **kwargs):
        """根据模块名和类别获取模块，是未初始化的"""
        # print(args, kwargs, cls)
        ps = self.ps(**kwargs)
        # 输入的名字也可以是ID
        exist_names = [x.split()[2] for x in ps.split('\n')[1::]]
        if name not in exist_names:
            exist_ids = [x.split()[0] for x in ps.split('\n')[1::]]
            assert len(exist_ids) == len(exist_names)
            if name in exist_ids:
                name = exist_names[exist_ids.index(name)]
            else:
                raise ModuleNotFoundError(f"Expected module name: '{name}' not found in exist modules: {exist_names}.")

        try:
            cls = cls if cls else self._name2cls[name]
        except Exception as e:
            raise ModuleNotFoundError(
                f'Error: {e}. \nAll modules: {self.modules}.')

        # print(cls)
        if cls == 'MODULE':
            return self.modules.get(name)
        elif cls == 'SCRIPT':
            return self.scripts.get(name)
        elif cls == 'STREAM':
            return self.streams.get(name)
        elif cls == 'IO':
            return self.ios.get(name)
        else:
            raise NotImplementedError(f'UAII get module name: {name} cls: {cls} not implemented.')

    def build_stream(self, cfg=None, **kwargs):
        """
        根据流配置文件创建新的流，注册到streams里
        cfg: stream config, for example：
            cfg = dict(
                type='radar_detection_stream',
                description='雷达目标探测流',
                models=[
                    dict(type='radardet', cfg=None)
            ])
        return: stream object
        """

        name = cfg['type']
        if name in self.streams.names:
            logger.warn(f'Build stream, name:{name} exist, return existing stream.')
            return self.get_stream(name=name)

        self.streams.build_new_stream(cfg, **kwargs)

        return self.get_stream(name)

    def get_stream(self, name):
        """获得存在的流，并初始化内部的所有模块"""
        stream = self.get_module(name=name, cls='STREAM')
        return stream

    def get_model(self, name):
        """读取单独的模块，内部是把单独的模块组装成流"""
        exists_streams = self.streams.names
        if name in exists_streams:
            mono_flow = self.get_module(name=name, cls='STREAM')
        else:
            cfg = dict(
                type=f'{name}_flow',
                description=f'mono module flow: {name}',
                models=[
                    dict(type=name, cfg=None),
                    ]
            )
            mono_flow = self.build_stream(cfg=cfg)
        return mono_flow

    def stream_info(self, name):
        return self.get_stream(name=name).info()

    def ps(self, stream=None, module=None, io=None, script=None):
        """返回全部模块的状态"""
        ps0 = [['ID', 'TYPE', 'NAME', 'STATUS', 'INCLUDE', 'DESCRIPTION']]
        if stream is None and module is None and io is None and script is None:
            tmp = self.streams.ps(include_dict=self.modules.name2id_dict)
            ps0 += tmp
            tmp2 = self.modules.ps(include_dict=self.ios.name2id_dict)
            ps0 += tmp2
            ps0 += self.ios.ps()
            # print(ps0, len(ps0))
            # ps0.append(self.scripts.ps())
        else:
            if stream:
                tmp = self.streams.ps(include_dict=self.modules.name2id_dict)
                ps0 += tmp
            if module:
                tmp = self.modules.ps(include_dict=self.modules.name2id_dict)
                ps0 += tmp
            if io:
                ps0 += self.ios.ps(module=None)
            if script:
                ps0 += self.scripts.ps(module=None)
        # ps_all = ps0
        # lenth = np.array([[len(x) for x in xx] for xx in ps_all])  # (n, 4)
        # print()
        # lenth = np.max(lenth, axis=0)  # (4,)
        # ps_list = [[f'{x:<{lenth[i]}}' for i, x in enumerate(xx)] for xx in ps_all]
        # ps_str = '\n'.join(['  '.join(x) for x in ps_list])
        ps_str = dm.misc.list2table(ps0)
        return ps_str

    def show_vis(self, ret, im0, target_names):

        return general.single_plot(ret, im0, target_names=target_names)




from .base_queue import AbstractQue


class AbstractInput(object):
    name = 'default_input_name'
    status = 'stopped'
    description = 'default_input_description'

    def __init__(self, *args, **kwargs):
        que = kwargs.pop('que', None)
        if que:
            maxlen = que.get('maxlen', 5)
            self.que = AbstractQue(maxlen, *args, **kwargs)
        else:
            self.que = None



import os
import cv2
from pathlib import Path
import numpy as np
import damei as dm
from copy import deepcopy

import torch.backends.cudnn as cudnn
from .datasets import LoadStreams, LoadImages


class Dataloader(object):
	def __init__(self, source=None, imgsz=None):
		self.video_suffix = ['.mp4', '.avi', '.wmv']
		self.imgsz = imgsz

		if isinstance(source, str):
			self.dataset = self.get_dataset(source)
		elif isinstance(source, list):
			datasets = [None] * len(source)
			for i, s in enumerate(source):
				datasets[i] = self.get_dataset(s)
			self.dataset = MultiDatasets(datasets)
		else:
			raise NotImplementedError

	def get_dataset(self, source):
		imgsz = self.imgsz
		video_suffix = self.video_suffix
		webcam = source.isnumeric() or source.startswith(('rtsp://', 'rtmp://', 'http://')) or source.endswith('.txt')
		if webcam:
			cudnn.benchmark = True  # set True to speed up constant image size inference
			dataset = LoadStreams(source, img_size=imgsz, flip=False)
			dataset.is_camera = True
			dataset.is_video = False
		else:
			dataset = LoadImages(source, img_size=imgsz)
			dataset.is_camera = False
			dataset.is_video = True if Path(source).suffix in video_suffix else False
		return dataset


class MultiDatasets(object):

	def __init__(self, datasets):
		self.datasets = datasets
		# print(f'multi {self.datasets}')
		self.is_camera = False
		self.is_video = False
		self.count = -1

	def __len__(self):
		datasets = self.datasets
		return int(np.mean([len(x) for x in datasets]))

	def __iter__(self):
		return self

	def __next__(self):
		datasets = self.datasets
		num = len(self.datasets)
		self.count += 1
		imgs = [None] * num
		im0s = [None] * num
		paths = [None] * num
		vid_caps = [None] * num
		for i, dset in enumerate(datasets):
			# 不同的数据集长度不一样呀可能会报错
			path, imgs[i], im0s[i], vid_caps[i] = next(dset)
			if dset.is_camera:
				im0s = im0s[0]
				path = f'{path[0]}/{self.count:0>6}.jpg'
			elif dset.is_video:
				path = f'{path}/{self.count:0>6}.jpg'
			else:
				pass
			paths[i] = path
			# print('\nxx', path)
		imgs = np.array(imgs)
		im0s = np.array(im0s)
		# print('xxx', imgs.shape, im0s.shape)
		return paths, imgs, im0s, vid_caps


def load_single_img(file, img_sz=640, return_list=False):
	if not os.path.exists(file):
		return
	im0 = cv2.imread(file)
	img, ratio, (dw, dh), recover = dm.general.letterbox(deepcopy(im0), new_shape=img_sz)
	img = img[:, :, ::-1].transpose(2, 0, 1)
	img = np.ascontiguousarray(img)

	if return_list:
		return [file], img[np.newaxis, :], [im0]
	else:
		return file, img, im0







from .registy import Registry

# print('registry init')

MODULES = Registry('module', build_func=None)
SCRIPTS = Registry('script', build_func=None)
# STREAMS = Registry('stream', build_func=None)
IOS = Registry('io', build_func=None, id_prefix='IO')




"""
模块的io，模块有输入流和输入流，每个流都有队列
"""
import os, sys
import time
import damei as dm
import time
# from ..registry import MODULES, SCRIPTS, IOS
from damei.nn.api.registry import MODULES, SCRIPTS, IOS
from damei.nn.uaii.stream.base_output import AbstractOutput


@IOS.register_module(name='uaii_output')
class UaiiOutput(AbstractOutput):
    name = 'uaii_output'
    status = 'stopped'
    description = 'defualt uaii output'

    def __init__(self, m_cfg, *args, **kwargs):
        maxlen = kwargs.get('maxlen', 5)
        super(UaiiOutput, self).__init__(maxlen=maxlen, *args, **kwargs)

@IOS.register_module(name='uaii_yolov5_output')
class SEYOLOv5OutStream(AbstractOutput):
    name = 'uaii_yolov5_output'
    status = 'stopped'
    description = 'seyolov5输出流'

    def __init__(self, m_cfg, *args, **kwargs):
        print(f'mcfg: {m_cfg}')
        self.m_cfg = m_cfg  # m_cfg是所有该模块的config，输出流只是该模块的一部分
        self.cfg = self.m_cfg['output_stream']
        # print(self.cfg, self.cfg.keys())
        maxlen = self.cfg['que'].get('maxlen', 5)
        super(SEYOLOv5OutStream, self).__init__(maxlen=maxlen, *args, **kwargs)
        self.save_dir = self.cfg.get('save_dir', None)
        self.save_txt = self.cfg.get('save_txt', False)
        self.save_conf = self.cfg.get('save_conf', False)
        self.print_ret = self.cfg.get('print_ret', True)
        self.que_wait = self.cfg['que'].get('wait', True)
        self.status = 'ready'

    def __repr__(self):
        format_str = f'<class "{self.__class__.__name__}"> ' \
                     f'(name="{self.name}", status="{self.status}", description="{self.description}")'
        return format_str

    def __call__(self, idx, ret, paths, imgs, im0s, *args, **kwargs):
        self.status = 'running'

        # 1.保存
        if self.save_dir:
            # TODO: 保存检测后的图像
            if self.save_txt:
                pass
                if self.save_conf:
                    pass
            raise NotImplementedError('save not implemented')

        # 2.打印结果
        if self.print_ret:
            self.analyse_and_print_ret(ret, names=self.m_cfg.model.names)

        # 3.输出到队列
        # TODO wait，现在是不wait
        if self.cfg['que']['enabled']:
            # self.status = ''
            self.push([idx, ret, paths, imgs, im0s], que_wait=self.que_wait)

    def analyse_and_print_ret(self, ret, names=None):
        """
        ret: list, [bs]个元素，每个元素是(num_obj, 13), 13=10+3, 2: status_idx, status_name, status_score
        """
        print(f'{"-":-<30} {"Print Detection Results":^25} {"-":-<30}')
        names = names if names else self.m_cfg.model.names
        head = ['Batch', 'Target', 'Bounding_box', 'Conf', 'Class', 'Class']
        for i in range(len(ret)):
            ptn_list = [[head]]
            single_ret = ret[i]  # 单图像的
            if single_ret is None:
                ptn_list.append([i, None, None, None, None])
            else:
                sp = single_ret.shape  # shape
                num_targets = sp[0]
                for j, target in enumerate(single_ret):
                    bbox_xyxy = target[:4]
                    conf = target[4]
                    target_cls = target[5]

                    tmp = [i+1,
                           f'{j+1}/{num_targets}',
                           ' '.join([str(int(x)) for x in bbox_xyxy]),
                           float(conf),
                           int(target_cls),
                           names[int(target_cls)],
                           ]

                    ptn_list.append(tmp)

            format_str = dm.misc.list2table(ptn_list, alignment="<")
            print(format_str)

        print(f'{"-":-<30} {"End Print":^25} {"-":-<30}')


@IOS.register_module(name='uaii_deepsort_output')
class DeepSortOutStream(AbstractOutput):
    name = 'uaii_deepsort_output'
    status = 'stopped'
    description = 'deepsort输出流'

    def __init__(self, m_cfg, *args, **kwargs):
        self.m_cfg = m_cfg  # m_cfg是所有该模块的config，输出流只是该模块的一部分
        self.cfg = self.m_cfg.output_stream
        # print(self.cfg, self.cfg.keys())
        maxlen = self.cfg.que.get('maxlen', 5)
        super(DeepSortOutStream, self).__init__(maxlen=maxlen, *args, **kwargs)
        self.que_wait = self.cfg.que.get('wait', True)

    def __repr__(self):
        format_str = f'{self.__class__.__name__}' \
                     f'(name="{self.name}")'
        return format_str

    def __call__(self, data, *args, **kwargs):
        if self.cfg.que.enabled:
            self.push(data, que_wait=self.que_wait)

